# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import ChatMessage
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.language_models.language_model import StreamingIterator
from synalinks.src.ops.operation import Operation
from synalinks.src.saving import serialization_lib


class Predict(Operation):
    """Perform a prediction using a `LanguageModel`."""

    def __init__(
        self,
        schema=None,
        data_model=None,
        language_model=None,
        streaming=False,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        if not schema and data_model:
            schema = data_model.get_schema()
        self.schema = schema
        self.data_model = data_model
        self.language_model = language_model
        if schema and streaming:
            streaming = False
        self.streaming = streaming
        self.lm_kwargs = kwargs

    async def call(self, x):
        value = await self.language_model(
            x,
            schema=self.schema,
            streaming=self.streaming,
            **self.lm_kwargs,
        )
        if isinstance(value, StreamingIterator):
            return value
        if not value:
            return None
        if self.schema:
            return JsonDataModel(json=value, schema=self.schema, name=self.name)
        else:
            return JsonDataModel(
                json=value, schema=ChatMessage.get_schema(), name=self.name
            )

    async def compute_output_spec(self, x):
        if self.schema:
            return SymbolicDataModel(schema=self.schema, name=self.name)
        else:
            return SymbolicDataModel(schema=ChatMessage.get_schema(), name=self.name)

    def get_config(self):
        config = {
            "schema": self.schema,
            "streaming": self.streaming,
            "name": self.name,
            "description": self.description,
        }
        language_model_config = serialization_lib.serialize_synalinks_object(
            self.language_model
        )
        config.update({"lm_kwargs": self.lm_kwargs})
        return {"language_model": language_model_config, **config}

    @classmethod
    def from_config(cls, config):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model")
        )
        lm_kwargs = config.pop("lm_kwargs")
        return cls(language_model=language_model, **config, **lm_kwargs)


@synalinks_export(["synalinks.ops.predict", "synalinks.ops.json.predict"])
async def predict(
    x,
    schema=None,
    data_model=None,
    language_model=None,
    streaming=False,
    name=None,
    description=None,
    **kwargs,
):
    """Perform a prediction using a `LanguageModel`.

    Predict consist in predicting a target data_model from an input data_model.
    This function uses a backend DataModel to get the target schema.

    Args:
        x (JsonDataModel | SymbolicDataModel): the input data model.
        data_model (DataModel): The target data model.
        language_model (LanguageModel): The language model to use
        streaming (bool): Enable streaming if True (Default to False)
        name (str): Optional. The name of the operation.
        description (str): Optional. The description of the operation.
        **kwargs (keyword arguments): Additional keywords forwarded to the
            LanguageModel call.

    Returns:
        (JsonDataModel | SymbolicDataModel): The resulting data model.
    """
    if language_model is None:
        raise ValueError("You should provide the `language_model` argument")
    if any_symbolic_data_models(x):
        return await Predict(
            schema=schema,
            data_model=data_model,
            language_model=language_model,
            streaming=False,
            name=name,
            description=description,
        ).symbolic_call(x)
    return await Predict(
        schema=schema,
        data_model=data_model,
        language_model=language_model,
        streaming=streaming,
        name=name,
        description=description,
        **kwargs,
    )(x)



# Modified from: keras/src/layers/layer_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import backend
from synalinks.src import modules
from synalinks.src import testing


class ModuleTest(testing.TestCase):
    async def test_compute_output_spec(self):
        class Query(backend.DataModel):
            query: str

        # Case: single output
        class TestModule(modules.Module):
            async def call(self, x):
                assert False  # Should never be called.

            async def compute_output_spec(self, inputs):
                return backend.SymbolicDataModel(data_model=inputs)

        module = TestModule()
        self.assertEqual(
            (await module(backend.SymbolicDataModel(data_model=Query))).get_schema(),
            backend.standardize_schema(Query.get_schema()),
        )

        # Case: tuple output
        class TestModule(modules.Module):
            async def call(self, x):
                assert False  # Should never be called.

            async def compute_output_spec(self, inputs):
                return (
                    backend.SymbolicDataModel(data_model=inputs),
                    backend.SymbolicDataModel(data_model=inputs),
                )

        module = TestModule()
        out = await module(backend.SymbolicDataModel(data_model=Query))
        self.assertIsInstance(out, tuple)
        self.assertEqual(len(out), 2)
        self.assertEqual(
            out[0].get_schema(), backend.standardize_schema(Query.get_schema())
        )
        self.assertEqual(
            out[1].get_schema(), backend.standardize_schema(Query.get_schema())
        )

        # Case: list output
        class TestModule(modules.Module):
            async def call(self, x):
                assert False  # Should never be called.

            async def compute_output_spec(self, inputs):
                return [
                    backend.SymbolicDataModel(data_model=inputs),
                    backend.SymbolicDataModel(data_model=inputs),
                ]

        module = TestModule()
        out = await module(backend.SymbolicDataModel(data_model=Query))

        self.assertIsInstance(out, list)
        self.assertEqual(len(out), 2)
        self.assertEqual(
            out[0].get_schema(), backend.standardize_schema(Query.get_schema())
        )
        self.assertEqual(
            out[1].get_schema(), backend.standardize_schema(Query.get_schema())
        )

        # Case: dict output
        class TestModule(modules.Module):
            async def call(self, x):
                assert False  # Should never be called.

            async def compute_output_spec(self, inputs):
                return {
                    "1": backend.SymbolicDataModel(data_model=inputs),
                    "2": backend.SymbolicDataModel(data_model=inputs),
                }

        module = TestModule()
        out = await module(backend.SymbolicDataModel(data_model=Query))

        self.assertIsInstance(out, dict)
        self.assertEqual(len(out), 2)
        self.assertEqual(
            out["1"].get_schema(), backend.standardize_schema(Query.get_schema())
        )
        self.assertEqual(
            out["2"].get_schema(), backend.standardize_schema(Query.get_schema())
        )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.language_models import deserialize as deserialize
from synalinks.src.language_models import serialize as serialize
from synalinks.src.language_models.language_model import LanguageModel as LanguageModel



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks import modules
from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules import Generator
from synalinks.src.modules import Input
from synalinks.src.programs import Program


class GeneratorModuleTest(testing.TestCase):
    def test_format_message(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        msgs = Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        ).format_messages(
            inputs=Query(query="What is the french city of aerospace and robotics?")
        )
        print(msgs.prettify_json())
        self.assertTrue(len(msgs.messages) == 2)

    def test_format_message_with_instructions(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama_chat/deepseek-r1")

        msgs = Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
            instructions="You are an helpfull assistant",
        ).format_messages(
            Query(query="What is the french city of aerospace and robotics?")
        )
        print(msgs.prettify_json())
        self.assertTrue(len(msgs.messages) == 2)

    def test_format_message_with_examples(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama_chat/deepseek-r1")

        msgs = Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
            examples=[
                (
                    {"query": "What is the capital of France?"},
                    {
                        "rationale": "The capital of France is well known",
                        "answer": "Paris",
                    },
                )
            ],
        ).format_messages(
            Query(query="What is the french city of aerospace and robotics?")
        )
        print(msgs.prettify_json())
        self.assertTrue(len(msgs.messages) == 2)

    def test_format_message_with_examples_and_instructions(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama_chat/deepseek-r1")

        msgs = Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
            examples=[
                (
                    {
                        "query": "What is the capital of France?",
                    },
                    {
                        "rationale": "The capital of France is well known",
                        "answer": "Paris",
                    },
                )
            ],
            instructions="You are an helpfull assistant",
        ).format_messages(
            Query(query="What is the french city of aerospace and robotics?")
        )
        print(msgs.prettify_json())
        self.assertTrue(len(msgs.messages) == 2)

    @patch("litellm.acompletion")
    async def test_basic_functional_setup(self, mock_completion):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama_chat/deepseek-r1")

        expected_string = (
            """{"rationale": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)
        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
        )

        result = await program(
            Query(query="What is the french city of aerospace and robotics?")
        )

        self.assertEqual(result.get_json(), json.loads(expected_string))

    @patch("litellm.acompletion")
    async def test_basic_functional_setup_with_schema(self, mock_completion):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama/mistral")

        expected_string = (
            """{"rationale": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        x0 = Input(data_model=Query)
        x1 = await Generator(
            schema=AnswerWithRationale.get_schema(),
            language_model=language_model,
        )(x0)
        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
        )

        result = await program(
            Query(query="What is the french city of aerospace and robotics?")
        )

        self.assertEqual(result.get_json(), json.loads(expected_string))

    @patch("litellm.acompletion")
    async def test_basic_subclassing_setup(self, mock_completion):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama/mistral")

        expected_string = (
            """{"rationale": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        class ChainOfThought(Program):
            def __init__(self, language_model):
                super().__init__()
                self.answer = Generator(
                    data_model=AnswerWithRationale, language_model=language_model
                )

            async def call(self, inputs):
                x = await self.answer(inputs)
                return x

        program = ChainOfThought(language_model=language_model)

        result = await program(
            Query(query="What is the french city of aerospace and robotics?")
        )

        self.assertEqual(result.get_json(), json.loads(expected_string))

    def test_serialization(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama/mistral")

        generator = Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )

        serialized_dict = modules.serialize(generator)
        new_generator = modules.deserialize(serialized_dict)
        # check that the nested object are good
        self.assertEqual(
            str(new_generator.language_model),
            str(generator.language_model),
        )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)


from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import is_observability_enabled
from synalinks.src.hooks.hook import Hook
from synalinks.src.hooks.logger import Logger
from synalinks.src.hooks.monitor import Monitor


@synalinks_export("synalinks.hooks.HookList")
class HookList(Hook):
    """Container abstracting a list of hooks.

    Container for `Hook` instances.

    This object wraps a list of `Hook` instances, making it possible
    to call them all at once via a single endpoint
    (e.g. `hook_list.on_module_begin(...)`).

    Args:
        hooks (list): List of `Hook` instances.
        add_logger (bool): Whether a `Logger` hook should be added, if one
            does not already exist in the `hooks` list.
        add_observability (bool): Whether a `Observability` hook should be added, if one
            does not already exist in the `hooks` list.
        module (Module): The `Module` these callbacks are used with.
        **params (keyword arguments): If provided, parameters will be passed to each
            `Hook` via `Hook.set_params`.
    """

    def __init__(
        self,
        hooks=None,
        add_logger=True,
        module=None,
        **params,
    ):
        self.hooks = tree.flatten(hooks) if hooks else []
        self._add_default_hooks(add_logger)
        self.set_module(module)
        self.set_params(params)

    def add_hook(self, hook):
        self.hooks.append(hook)

    def set_params(self, params):
        self.params = params
        if params:
            for hook in self.hooks:
                hook.set_params(params)

    def _add_default_hooks(self, add_logger):
        self._logger = None
        self._monitor = None

        for hook in self.hooks:
            if isinstance(hook, Logger):
                self._logger = hook

        if self._logger is None and add_logger:
            self._logger = Logger()
            self.hooks.append(self._logger)

        if self._monitor is None and is_observability_enabled():
            self._monitor = Monitor()
            self.hooks.append(self._monitor)

    def set_module(self, module):
        if not module:
            return
        super().set_module(module)
        for hook in self.hooks:
            hook.set_module(module)

    def on_call_begin(
        self,
        call_id,
        inputs=None,
    ):
        for hook in self.hooks:
            hook.on_call_begin(
                call_id,
                inputs=inputs,
            )

    def on_call_end(
        self,
        call_id,
        outputs=None,
        exception=None,
    ):
        for hook in self.hooks:
            hook.on_call_end(
                call_id,
                outputs=outputs,
                exception=exception,
            )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import List

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import Trainable
from synalinks.src.modules.module import Module
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.saving import serialization_lib


class Step(DataModel):
    """The individual step to execute"""
    step: str = Field(
        description="The step to execute",
    )


class SequentialPlan(Trainable):
    """The sequential step by step plan to achieve the task"""
    steps: List[str] = Field(
        description="The list of steps",
    )


class SequentialPlanSynthesis(Module):
    """A module that executes a sequential plan of steps.

    This module features a sequential plan as a trainable variable, allowing optimizers
    to refine the plan during the training loop based on iterative feedback.
    
    Basically learning to plan based on iterative feedback and automatic selection of the best plan.

    The module executes each step in the plan sequentially, passing the output of each
    step as input to the next step. The runner is responsible for executing
    each individual step. The most common runners are usually a `FunctionCallingAgent`, 
    `ChainOfThought` or `Generator` module, but you can use any Module or Program.
    
    This module start by defaut without any plan, so it is equivalent to a `ChainOfThought` module,
    iteratively, the plan will be constructed and optimized to solve the task.
    
    This module works **ONLY** with advanced optimizers (**NOT** the `RandomFewShot` optimizer).
    
    **Note**: The inputs are forwarded to the runner each time by concatenating the inputs with
    the previous steps outputs. So **ensure that the runner doesn't returns the inputs**, use 
    `return_inputs=False` or `return_inputs_with_trajectory=False` when configuring your runner.

    Example:
    
    ```python
    import synalinks
    import asyncio
    
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )
        
    class FinalReport(synalinks.DataModel):
        report: str = synalinks.Field(
            description="The final report",
        )
    
    class TaskSummary(synalinks.DataModel):
        summary: str = synalinks.Field(
            description="The summary of the executed task",
        )
    
    async def main():
        tools = # ... tools definition (see `FunctionCallingAgent`)
        
        inputs = synalinks.Input(data_model=Query)
        outputs = await synalinks.SequentialPlanSynthesis(
            data_model=FinalReport,
            language_model=language_model,
            runner=synalinks.FunctionCallingAgent(
                data_model=TaskSummary,
                language_model=language_model,
                tools=tools,
                return_inputs_with_trajectory=False,
            ),
        )(inputs)
        
        program = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="planner_agent",
            description="An agent that learn a step by step plan to achieve a task",
        )
        
    ```
    
    Args:
        schema (dict): The target JSON schema.
            If not provided use the `data_model` to infer it.
        data_model (DataModel | SymbolicDataModel | JsonDataModel): The target data
            model for structured output.
        language_model (LanguageModel): The language model to use.
        steps (list): Optional. The default list of steps being a list of strings.
        seed_steps (list): Optional. A list of steps to use as seed for the
            optimization. If not provided, use the default steps as seed.
        runner (Module | Program): Required. The runner that executes each step.
        return_inputs (bool): Optional. Whether or not to concatenate the inputs to
            the outputs (Default to False).
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.    
    """
    
    def __init__(
        self,
        schema=None,
        data_model=None,
        language_model=None,
        steps=None,
        seed_steps=None,
        runner=None,
        return_inputs=True,
        name=None,
        description=None,
        trainable=True,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        if not schema and data_model:
            schema = data_model.get_schema()
        self.schema = schema
        
        if not steps:
            steps = []
        self.steps = steps
            
        if not seed_steps:
            seed_steps = [[]]
        
        self.seed_steps = seed_steps
        if not runner:
            raise ValueError("The `runner` parameter is required.")
        if not isinstance(runner, Module):
            raise ValueError("The `runner` parameter should be a `Module` or `Program`.")
        
        self.language_model = language_model
        self.runner = runner
        self.return_inputs = return_inputs
        
        self.state = self.add_variable(
            initializer=SequentialPlan(
                steps=self.steps,
                seed_candidates=self.seed_steps,
            ).get_json(),
            data_model=SequentialPlan,
            name=self.name+"_state",
        )
        
        self.final_generator = ChainOfThought(
            schema=self.schema,
            language_model=self.language_model,
            return_inputs=self.return_inputs,
            name=self.name + "_final_generator",
        )
        
    async def call(self, inputs, training=False):
        steps = self.state.get("steps")
        previous_steps = None
        for i, step in enumerate(steps):
            step_result = await self.runner(inputs, training=training)
            if not previous_steps:
                previous_steps = step_result
            else:
                previous_steps = await ops.concat(
                    previous_steps,
                    step_result,
                    name=self.name+f"_step_{i}_with_inputs",
                )
            inputs = await ops.concat(
                inputs,
                await ops.concat(
                    previous_steps,
                    Step(step=step),
                    name=self.name+f"_step_{i}",
                ),
                name=self.name+f"_step_{i}_with_inputs",
            )
        return await self.final_generator(inputs, training=training)
        
    async def compute_output_spec(self, inputs, training=False):
        _ = await self.runner(inputs)
        return await self.final_generator(inputs)
        
    def get_config(self):
        config = {
            "schema": self.schema,
            "steps": self.steps,
            "seed_steps": self.seed_steps,
            "return_inputs": self.return_inputs,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model,
            )
        }
        runner_config = {
            "runner": serialization_lib.serialize_synalinks_object(
                self.runner,
            )
        }
        return {
            **config,
            **language_model_config,
            **runner_config,
        }

    @classmethod
    def from_config(cls, config):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model"),
        )
        runner = serialization_lib.deserialize_synalinks_object(
            config.pop("runner"),
        )
        return cls(
            language_model=language_model,
            runner=runner,
            **config,
        )


# Modified from: keras/src/losses/losses.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend.common import numpy as np
from synalinks.src.rewards.reward import Reward
from synalinks.src.rewards.reward import squeeze_or_expand_to_same_rank
from synalinks.src.rewards.reward_wrappers import RewardFunctionWrapper


@synalinks_export("synalinks.rewards.cosine_similarity")
async def cosine_similarity(y_true, y_pred, embedding_model=None, axis=-1):
    """
    Computes the cosine similarity between `y_true` and `y_pred`.

    Formula:

    ```
    reward = (sum(l2_norm(y_true) * l2_norm(y_pred))+1) / 2
    ```

    The formula is similar to the classic cosine similarity used in deep learning,
    but scaled to [0.0, 1.0] and adjusted to have a reward that tend
    towards 1.0 if the two objects are similar (and 0.0 otherwise).

    Args:
        y_true (JsonDataModel): The ground truth JSON data_model.
        y_pred (JsonDataModel): The predicted JSON data_model.
        embedding_model (EmbeddingModel): The embedding model to use to compute the
            cosine similarity.
        axis (int): (Optional) Defaults to `-1`. The dimension along which the cosine
            similarity is computed.

    Returns:
        (float): The reward value, which tend to 1.0 if the values are similar,
            and towards 0.0 otherwise.
    """
    reward = 0.0
    if y_pred is not None:
        y_true = await ops.embedding(y_true, embedding_model=embedding_model)
        y_pred = await ops.embedding(y_pred, embedding_model=embedding_model)
        y_true = np.convert_to_tensor(y_true.get("embeddings"))
        y_pred = np.convert_to_tensor(y_pred.get("embeddings"))
        y_true, y_pred = squeeze_or_expand_to_same_rank(y_true, y_pred)
        y_pred = np.normalize(y_pred, axis=axis)
        y_true = np.normalize(y_true, axis=axis)
        reward = (np.sum(y_true * y_pred, axis=axis) + 1) / 2
    return reward


@synalinks_export(
    [
        "synalinks.CosineSimilarity",
        "synalinks.rewards.CosineSimilarity",
    ]
)
class CosineSimilarity(RewardFunctionWrapper):
    """
    Computes the cosine similarity between `y_true` and `y_pred`.

    Formula:

    ```
    reward = (sum(l2_norm(y_true) * l2_norm(y_pred))+1) / 2
    ```

    The formula is similar to the classic cosine similarity used in deep learning,
    but scaled to [0.0, 1.0] and adjusted to have a reward that tend
    towards 1.0 if the two objects are similar (and 0.0 otherwise).

    Example:

    ```python
    program.compile(
        reward=synalinks.rewards.CosineSimilarity(
            embedding_model=embedding_model
        )
        optimizer=synalinks.optimizers.RandomFewShot(),
    )
    ```

    Args:
        embedding_model (EmbeddingModel): The embedding model to use to compute the
            cosine similarity.
        axis (int): (Optional) Defaults to `-1`. The dimension along which the cosine
            similarity is computed.
        name (str): (Optional) string name of the reward instance.
        in_mask (list): (Optional) list of keys to keep to compute the reward.
        out_mask (list): (Optional) list of keys to remove to compute the reward.
    """

    def __init__(
        self,
        embedding_model=None,
        axis=-1,
        name="cosine_similarity",
        in_mask=None,
        out_mask=None,
    ):
        super().__init__(
            fn=cosine_similarity,
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
            axis=axis,
            embedding_model=embedding_model,
        )

    def get_config(self):
        config = Reward.get_config()
        from synalinks.src.saving.serialization_lib import serialize_synalinks_object

        embedding_model_config = {
            "embedding_model": serialize_synalinks_object(self.embedding_model)
        }
        return {**config, **embedding_model_config}

    @classmethod
    def from_config(cls, config):
        from synalinks.saving.serialization_lib import deserialize_synalinks_object

        embedding_model = deserialize_synalinks_object(config.pop("embedding_model"))
        return cls(embedding_model=embedding_model, **config)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.modules.module import Module


@synalinks_export(
    [
        "synalinks.And",
        "synalinks.modules.And",
    ]
)
class And(Module):
    """Perform a logical And operation.

    It takes as input a list of data models,
    and returns a concatenation of them.

    If any input is None, then it output None.

    Table:

    | `x1`   | `x2`   | Logical And (`&`) |
    | ------ | ------ | ----------------- |
    | `x1`   | `x2`   | `x1 + x2`         |
    | `x1`   | `None` | `None`            |
    | `None` | `x2`   | `None`            |
    | `None` | `None` | `None`            |

    Args:
        **kwargs (keyword arguments): Standard keyword arguments for the module.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def call(self, inputs, training=False):
        output = inputs[0]
        for i in range(1, len(inputs)):
            output = await ops.logical_and(
                output,
                inputs[i],
                name=self.name + f"_module_and_{i}",
            )
        return output
